@@ -1,12 +1,11 @@
 #include <Python.h>
 #include <ctype.h>
 #include <datetime.h>
+#include "timezone.h"
 
 #define STRINGIZE(x)            #x
 #define EXPAND_AND_STRINGIZE(x) STRINGIZE(x)
 
-#define PY_VERSION_AT_LEAST_32 \
-    ((PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION >= 2) || PY_MAJOR_VERSION > 3)
 #define PY_VERSION_AT_LEAST_33 \
     ((PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION >= 3) || PY_MAJOR_VERSION > 3)
 #define PY_VERSION_AT_LEAST_36 \
@@ -14,12 +13,29 @@
 #define PY_VERSION_AT_LEAST_37 \
     ((PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION >= 7) || PY_MAJOR_VERSION > 3)
 
-#if !PY_VERSION_AT_LEAST_37
-static PyObject *fixed_offset;
+// PyPy compatibility for cPython 3.7's Timezone API was added to PyPy 7.3.6
+// https://foss.heptapod.net/pypy/pypy/-/merge_requests/826
+#ifdef PYPY_VERSION
+    #define SUPPORTS_37_TIMEZONE_API \
+            (PYPY_VERSION_NUM >= 0x07030600)
+#else
+    #define SUPPORTS_37_TIMEZONE_API \
+            PY_VERSION_AT_LEAST_37
 #endif
 
 static PyObject *utc;
 
+#if CISO8601_CACHING_ENABLED
+/* 2879 = (1439 * 2) + 1, number of offsets from UTC possible in
+ * Python (ie. [-1439, 1439]).
+ *
+ * 0 - 1438 = Negative offsets [-1439..-1]
+ * 1439 = Zero offset
+ * 1440 - 2878 = Positive offsets [1...1439]
+ */
+static PyObject *tz_cache[2879] = {NULL};
+#endif
+
 #define PARSE_INTEGER(field, length, field_name)                       \
     for (i = 0; i < length; i++) {                                     \
         if (*c >= '0' && *c <= '9') {                                  \
@@ -144,6 +160,9 @@ _parse(PyObject *self, PyObject *args, int parse_any_tzinfo, int rfc3339_only)
         usecond = 0;
     int time_is_midnight = 0;
     int tzhour = 0, tzminute = 0, tzsign = 0;
+#if CISO8601_CACHING_ENABLED
+    int tz_index = 0;
+#endif
     PyObject *delta;
     PyObject *temp;
     int extended_date_format = 0;
@@ -427,34 +446,47 @@ _parse(PyObject *self, PyObject *args, int parse_any_tzinfo, int rfc3339_only)
                 tzminute += 60 * tzhour;
                 tzminute *= tzsign;
 
-#if !PY_VERSION_AT_LEAST_32
-                if (fixed_offset == NULL || utc == NULL) {
-                    PyErr_SetString(PyExc_ImportError,
-                                    "Cannot parse a timestamp with time zone "
-                                    "information without the pytz dependency. "
-                                    "Install it with `pip install pytz`.");
-                    return NULL;
-                }
-#endif
-
                 if (tzminute == 0) {
                     tzinfo = utc;
                 }
-                else {
-#if PY_VERSION_AT_LEAST_37
-                    delta = PyDelta_FromDSU(0, 60 * tzminute, 0);
-                    tzinfo = PyTimeZone_FromOffset(delta);
+                else if (abs(tzminute) >= 1440) {
+                    /* Format the error message as if we were still using pytz
+                     * for Python 2 and datetime.timezone for Python 3.
+                     * This is done to maintain complete backwards
+                     * compatibility with ciso8601 2.0.x. Perhaps change to a
+                     * simpler message in ciso8601 v3.0.0.
+                     */
+#if PY_MAJOR_VERSION >= 3
+                    delta = PyDelta_FromDSU(0, tzminute * 60, 0);
+                    PyErr_Format(PyExc_ValueError,
+                                 "offset must be a timedelta"
+                                 " strictly between -timedelta(hours=24) and"
+                                 " timedelta(hours=24),"
+                                 " not %R.",
+                                 delta);
                     Py_DECREF(delta);
-#elif PY_VERSION_AT_LEAST_32
-                    tzinfo = PyObject_CallFunction(
-                        fixed_offset, "N",
-                        PyDelta_FromDSU(0, 60 * tzminute, 0));
 #else
-                    tzinfo =
-                        PyObject_CallFunction(fixed_offset, "i", tzminute);
+                    PyErr_Format(PyExc_ValueError,
+                                 "('absolute offset is too large', %d)",
+                                 tzminute);
 #endif
+                    return NULL;
+                }
+                else {
+#if CISO8601_CACHING_ENABLED
+                    tz_index = tzminute + 1439;
+                    if ((tzinfo = tz_cache[tz_index]) == NULL) {
+                        tzinfo = new_fixed_offset(60 * tzminute);
+
+                        if (tzinfo == NULL) /* ie. PyErr_Occurred() */
+                            return NULL;
+                        tz_cache[tz_index] = tzinfo;
+                    }
+#else
+                    tzinfo = new_fixed_offset(60 * tzminute);
                     if (tzinfo == NULL) /* ie. PyErr_Occurred() */
                         return NULL;
+#endif
                 }
             }
         }
@@ -473,6 +505,10 @@ _parse(PyObject *self, PyObject *args, int parse_any_tzinfo, int rfc3339_only)
     /* Make sure that there is no more to parse. */
     if (*c != '\0') {
         PyErr_Format(PyExc_ValueError, "unconverted data remains: '%s'", c);
+#if !CISO8601_CACHING_ENABLED
+        if (tzinfo != Py_None && tzinfo != utc)
+            Py_DECREF(tzinfo);
+#endif
         return NULL;
     }
 
@@ -480,8 +516,10 @@ _parse(PyObject *self, PyObject *args, int parse_any_tzinfo, int rfc3339_only)
         year, month, day, hour, minute, second, usecond, tzinfo,
         PyDateTimeAPI->DateTimeType);
 
+#if !CISO8601_CACHING_ENABLED
     if (tzinfo != Py_None && tzinfo != utc)
         Py_DECREF(tzinfo);
+#endif
 
     if (obj && time_is_midnight) {
         delta = PyDelta_FromDSU(1, 0, 0); /* 1 day */
@@ -542,12 +580,6 @@ PyInit_ciso8601(void)
 initciso8601(void)
 #endif
 {
-#if !PY_VERSION_AT_LEAST_32
-    PyObject *pytz;
-#elif !PY_VERSION_AT_LEAST_37
-    PyObject *datetime;
-#endif
-
 #if PY_MAJOR_VERSION >= 3
     PyObject *module = PyModule_Create(&moduledef);
 #else
@@ -558,28 +590,23 @@ initciso8601(void)
                                EXPAND_AND_STRINGIZE(CISO8601_VERSION));
 
     PyDateTime_IMPORT;
-#if PY_VERSION_AT_LEAST_37
-    utc = PyDateTime_TimeZone_UTC;
-#elif PY_VERSION_AT_LEAST_32
-    datetime = PyImport_ImportModule("datetime");
-    if (datetime == NULL)
-        return NULL;
-    fixed_offset = PyObject_GetAttrString(datetime, "timezone");
-    if (fixed_offset == NULL)
-        return NULL;
-    utc = PyObject_GetAttrString(fixed_offset, "utc");
-    if (utc == NULL)
+
+    // PyMODINIT_FUNC is void in Python 2, returns PyObject* in Python 3
+    if (initialize_timezone_code(module) < 0) {
+#if PY_MAJOR_VERSION >= 3
         return NULL;
 #else
-    pytz = PyImport_ImportModule("pytz");
-    if (pytz == NULL) {
-        PyErr_Clear();
-    }
-    else {
-        fixed_offset = PyObject_GetAttrString(pytz, "FixedOffset");
-        utc = PyObject_GetAttrString(pytz, "UTC");
-    }
+        return;
 #endif
+    }
+
+#if SUPPORTS_37_TIMEZONE_API
+    utc = PyDateTime_TimeZone_UTC;
+#else
+    utc = new_fixed_offset(0);
+#endif
+
+// PyMODINIT_FUNC is void in Python 2, returns PyObject* in Python 3
 #if PY_MAJOR_VERSION >= 3
     return module;
 #endif
